
#include <iostream>
#include <stdlib.h>
#include <math.h>

#include <SDL.h>
#include <SDL_thread.h>

#include "../pipes/streamPipe.h"
#include "streamElement.h"
#include "tone.h"

Tone::Tone(
	StreamPipe* in, StreamPipe* out, double _freq,
	double _linesPerSecond, double _gain,
	unsigned int _waveType, double gamma,
	double _minHz, double _maxHz
    ) : StreamElement( in, out ), freq( _freq ),
	linesPerSecond( _linesPerSecond ),
	gain( _gain ), waveType( _waveType ),
	minHz( _minHz ), maxHz( _maxHz )
{
    for ( unsigned int ii=0; ii < 256; ++ii ) {
	this->intensityMap[ ii ] = pow( (double)ii / 255.0, 1.0 / gamma );
    }
}

Tone::~Tone( void )
{
}

static double
triangleWave( double angle )
{
    double waveNumber;
    double proportion = modf( angle / ( 2.0 * M_PI ), &waveNumber );

    if ( proportion < 0.5 ) {
	return -proportion * 4.0 + 1.0;
    } else {
	return proportion * 4.0 - 3.0;
    }
}

static double
squareWave( double angle )
{
    double waveNumber;
    double proportion = modf( angle / ( 2.0 * M_PI ), &waveNumber );

    if ( proportion < 0.5 ) {
	return 1.0;
    } else {
	return -1.0;
    }
}

static double
sawtoothWave( double angle )
{
    double waveNumber;
    double proportion = modf( angle / ( 2.0 * M_PI ), &waveNumber );

    return proportion * 2.0 - 1.0;
}

static double
toothsawWave( double angle )
{
    double waveNumber;
    double proportion = modf( angle / ( 2.0 * M_PI ), &waveNumber );

    return -proportion * 2.0 + 1.0;
}

int
Tone::run( void )
{
    int rLen;
    unsigned int width;

    rLen = this->in->read( (unsigned char*)&width, sizeof( width ) );

    if ( rLen not_eq sizeof( width ) ) {
	std::cerr << "Tone couldn't read width" << std::endl;
	return __LINE__;
    }

    double* hzTable = new double[ width ];
    double logMinHz = log( this->minHz );
    double logMaxHz = log( this->maxHz );

    for ( unsigned int ii=0; ii < width; ++ii ) {
	double horz = (double) ii / (double)width;
	double logHz = logMinHz + horz * ( logMaxHz - logMinHz );

	hzTable[ ii ] = exp( logHz );
    }

    double realGain = gain / (double)width;

    unsigned char* buf[2];
    buf[ 0 ] = new unsigned char[ width ];
    buf[ 1 ] = new unsigned char[ width ];

    double (*waveForm)( double angle ) = sin;

    switch ( this->waveType ) {
    case TRIANGLE:
	waveForm = triangleWave;
	break;
    case SQUARE:
	waveForm = squareWave;
	break;
    case SAWTOOTH:
	waveForm = sawtoothWave;
	break;
    case TOOTHSAW:
	waveForm = toothsawWave;
	break;
    }

    rLen = this->in->read( buf[ 0 ], width );

    int previousRow = -1;
    unsigned int yy=0;

    while ( rLen == width ) {
	double vert = ( (double) yy * this->linesPerSecond )
	    / (double) this->freq;

	double realRow;
	double proportion = modf( vert, &realRow );
	int otherRow = ( (int)realRow ) % 2;
	int row = ( otherRow + 1 ) % 2;

	if ( row not_eq previousRow ) {
	    previousRow = row;
	    rLen = this->in->read( buf[ row ], width );
	    if ( rLen not_eq width ) {
		break;
	    }
	}

	unsigned char* aPtr = buf[ row ];
	unsigned char* bPtr = buf[ otherRow ];

	double angleFactor = (double) yy * 2.0 * M_PI / this->freq;
	double samp = 0.0;

	for ( unsigned int ii=0; ii < width; ++ii ) {
	    double partA = this->intensityMap[ *aPtr++ ];
	    double partB = this->intensityMap[ *bPtr++ ];

	    double intensity
		= proportion * partA + ( 1.0 - proportion ) * partB;

	    samp += intensity * (*waveForm)( hzTable[ ii ] * angleFactor );
	}

	samp *= realGain;

	this->out->write( &samp, sizeof( samp ) );
	++yy;
    }

    delete[] buf[1];
    delete[] buf[0];
    delete[] hzTable;

    return 0;
}
